@@ -226,6 +226,8 @@ struct ibmca_op_ctx {
                 int saltlen;
                 struct ibmca_pss_params pss;
                 EVP_MD_CTX *md_ctx;
+                unsigned char *signature;
+                size_t signature_len;
             } signature; /* For operation EVP_PKEY_OP_SIGN/VERIFY */
         } rsa; /* For type EVP_PKEY_RSA and EVP_PKEY_RSA_PSS */
         union {
@@ -233,12 +235,17 @@ struct ibmca_op_ctx {
                 int selection;
                 int curve_nid;
                 point_conversion_form_t format;
+                unsigned char *dhkem_ikm;
+                size_t dhkem_ikmlen;
             } gen; /* For operation EVP_PKEY_OP_KEYGEN */
             struct {
                 EVP_MD *md;
                 bool set_md_allowed;
                 size_t md_size;
                 EVP_MD_CTX *md_ctx;
+                unsigned int nonce_type;
+                unsigned char *signature;
+                size_t signature_len;
             } signature; /* For operation EVP_PKEY_OP_SIGN/VERIFY */
             struct {
                 struct ibmca_key *peer_key;
